@@ -19,9 +19,21 @@
 //! assert_eq!((mx * my).residue(), x * y % m);
 //! ```
 //!
+//! # Comparison of fast division / modular arithmetics
+//! Several fast division / modulo tricks are provided in these crate, the difference of them are listed below:
+//! - [PreInv]: pre-compute modular inverse of the divisor, only applicable to exact division
+//! - Barret (to be implemented): pre-compute (rational approximation of) the reciprocal of the divisor,
+//!     applicable to fast division and modulo
+//! - [Montgomery]: Convert the dividend into a special form by shifting and pre-compute a modular inverse,
+//!     only applicable to fast modulo, but faster than Barret reduction
+//! - [MersenneInt]: Specialization of modulo in form `2^P-K`
+//! 
 
-// XXX: consider implementing lookup table based modulo?
+// XXX: Other fast modular arithmetic tricks
+// REF: https://github.com/lemire/fastmod & https://arxiv.org/pdf/1902.01961.pdf
 // REF: https://eprint.iacr.org/2014/040.pdf
+// REF: https://github.com/ridiculousfish/libdivide/
+// REF: Faster Interleaved Modular Multiplication Based on Barrett and Montgomery Reduction Methods (work for modulus in certain form)
 
 #![no_std]
 #[cfg(any(feature = "std", test))]
@@ -114,6 +126,7 @@ pub trait ModularSymbols<Modulus = Self> {
 }
 
 // TODO: Discrete log aka index, follow the behavior of FLINT `n_discrete_log_bsgs`
+// REF: https://github.com/vks/discrete-log
 // fn logm(self, base: Modulus, m: Modulus);
 
 /// Collection of common modular arithmetic operations
@@ -169,16 +182,30 @@ pub trait ModularInteger:
     /// constructor to prevent unnecessary overhead of pre-computation.
     fn convert(&self, n: Self::Base) -> Self;
 
-    // Calculate the value of self + self
+    /// Calculate the value of self + self
     fn double(self) -> Self;
 
-    // Calculate the value of self * self
+    /// Calculate the value of self * self
     fn square(self) -> Self;
 }
-
 // XXX: implement this trait for ff::PrimeField?
+// TODO: implement invm_range (Modular inverse in certain range) and crt (Chinese Remainder Theorem), REF: bubblemath crate
+
+/// Utility function for exact division, with precomputed helper values
+/// 
+/// # Available Pre-computation types:
+/// - `()`: No pre-computation, the implementation relies on native integer division
+/// - [PreInv]: With Pre-computed modular inverse
+pub trait DivExact<Rhs, Precompute> : Sized {
+    type Output;
+
+    /// Check if d divides self with the help of the precomputation. If d divides self,
+    /// then the quotient is returned.
+    fn div_exact(self, d: Rhs, pre: &Precompute) -> Option<Self::Output>;
+}
 
 mod barret;
+mod preinv;
 mod double;
 mod mersenne;
 mod monty;
@@ -186,6 +213,7 @@ mod prim;
 
 pub use double::{udouble, umax};
 pub use mersenne::MersenneInt;
+pub use preinv::PreInv;
 pub use monty::{Montgomery, MontgomeryInt};
 
 #[cfg(feature = "num-bigint")]
